/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.jcip.annotations.Immutable;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Multimap;

/**
 * This class represents a context-free grammar.
 *
 * <p>
 * A context-free grammar is a formal grammar in which the left-hand side of
 * each production rule consists of only a single non-terminal symbol.
 *
 * <p>
 * A formal grammar is a set of rules for rewriting strings, along with a
 * "start symbol" from which rewriting must start.
 *
 * <p>
 * <i>For more information see the Wikipedia article: <a
 * href="http://en.wikipedia.org/wiki/Context-free_grammar">Context-free
 * grammar</a></i>
 *
 * @author Clemens Akens
 */
@Immutable
public final class Grammar {

    /**
     * This static field stores the object mapper to generate the JSON string
     * representation from the JSON object representation of a context-free
     * grammar.
     */
    private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

    /**
     * This static method creates a new context-free grammar.
     *
     * @param startSymbol
     *            The start symbol of the new context-free grammar.
     * @param rules
     *            The (possibly empty) set of all production rules of the new
     *            context-free grammar.
     * @return The new context-free grammar.
     */
    public static Grammar create(final NonTerminal startSymbol, final Collection<Rule> rules) {
        return new Grammar(startSymbol, ImmutableSet.copyOf(rules));
    }

    /**
     * This static method creates a new context-free grammar.
     *
     * @param startSymbol
     *            The start symbol of the new context-free grammar.
     * @param rules
     *            The (possibly empty) set of all production rules of the new
     *            context-free grammar.
     * @return The new context-free grammar.
     */
    public static Grammar create(final NonTerminal startSymbol, final Rule... rules) {
        return new Grammar(startSymbol, ImmutableSet.copyOf(rules));
    }

    /**
     * This field stores the set of all non-terminal symbols that occur on the
     * left-hand or right-hand side of each production rule of this context-free
     * grammar including the start symbol.
     */
    private final ImmutableSet<NonTerminal> nonTerminals;

    /**
     * This field stores the set of all production rules of this context-free
     * grammar.
     */
    private final ImmutableSet<Rule> rules;

    /**
     * This field stores the sets of all production rules of this context-free
     * grammar mapped by their respective left-hand sides.
     */
    private final ImmutableSetMultimap<NonTerminal, Rule> rulesByNonTerminal;

    /** This field stores the start symbol of this context-free grammar. */
    private final NonTerminal startSymbol;

    /**
     * This field stores the the set of all terminal symbols that occur on the
     * right-hand side of each production rule of this context-free grammar.
     */
    private final ImmutableSet<Terminal> terminals;

    /**
     * This constructor instantiates a new context-free grammar.
     *
     * @param startSymbol
     *            The start symbol of the new context-free grammar.
     * @param rules
     *            The (possibly empty) set of all production rules of the new
     *            context-free grammar.
     */
    private Grammar(final NonTerminal startSymbol, final ImmutableSet<Rule> rules) {
        final Set<NonTerminal> nonTerminals = new HashSet<NonTerminal>();
        final Multimap<NonTerminal, Rule> rulesByNonTerminal = HashMultimap.create();
        final Set<Terminal> terminals = new HashSet<Terminal>();

        nonTerminals.add(Preconditions.checkNotNull(startSymbol, "startSymbol is null"));

        for (final Rule rule : rules) {
            rulesByNonTerminal.put(rule.getLeftHandSide(), rule);
            nonTerminals.add(rule.getLeftHandSide());

            for (final Symbol symbol : rule.getRightHandSide()) {
                if (symbol.isNonTerminal()) {
                    nonTerminals.add((NonTerminal) symbol);
                } else {
                    terminals.add((Terminal) symbol);
                }
            }
        }

        this.nonTerminals = ImmutableSet.copyOf(nonTerminals);
        this.rules = rules;
        this.rulesByNonTerminal = ImmutableSetMultimap.copyOf(rulesByNonTerminal);
        this.startSymbol = startSymbol;
        this.terminals = ImmutableSet.copyOf(terminals);
    }

    /**
     * This method indicates whether some other object is "equal to" this one.
     *
     * @param object
     *            The other object.
     * @return True, if this object is equal to the other object.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(final Object object) {
        if (object == this) {
            return true;
        }

        if (object == null) {
            return false;
        }

        if (object.getClass() != this.getClass()) {
            return false;
        }

        final Grammar other = (Grammar) object;

        return Objects.equal(other.rules, this.rules) && Objects.equal(other.startSymbol, this.startSymbol);
    }

    /**
     * This method returns the set of all non-terminal symbols that occur on the
     * left-hand or right-hand side of each production rule of this context-free
     * grammar including the start symbol.
     *
     * @return The set of all non-terminal symbols that occur on the left-hand
     *         or right-hand side of each production rule of this context-free
     *         grammar including the start symbol.
     */
    public ImmutableSet<NonTerminal> getNonTerminals() {
        return this.nonTerminals;
    }

    /**
     * This method returns the set of all production rules of this context-free
     * grammar.
     *
     * @return The set of all production rules of this context-free grammar.
     */
    public ImmutableSet<Rule> getRules() {
        return this.rules;
    }

    /**
     * This method returns the sets of all production rules of this context-free
     * grammar mapped by their respective left-hand sides.
     *
     * @return The sets of all production rules of this context-free grammar
     *         mapped by their respective left-hand sides.
     */
    public ImmutableSetMultimap<NonTerminal, Rule> getRulesByNonTerminal() {
        return this.rulesByNonTerminal;
    }

    /**
     * This method returns the start symbol of this context-free grammar.
     *
     * @return The start symbol of this context-free grammar.
     */
    public NonTerminal getStartSymbol() {
        return this.startSymbol;
    }

    /**
     * This method returns the set of all terminal symbols that occur on the
     * right-hand side of each production rule of this context-free grammar.
     *
     * @return The set of all terminal symbols that occur on the right-hand side
     *         of each production rule of this context-free grammar.
     */
    public ImmutableSet<Terminal> getTerminals() {
        return this.terminals;
    }

    /**
     * This method returns the hash code value for this object.
     *
     * @return The hash code value for this object.
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return Objects.hashCode(this.rules, this.startSymbol);
    }

    /**
     * This method returns the JSON object representation of this context-free
     * grammar.
     *
     * @return The JSON object representation of this context-free grammar.
     */
    public Object toObject() {
        final Map<String, Object> object1 = new HashMap<String, Object>();

        if (this.startSymbol != null) { // EqualsVerifier: Non-nullity
            object1.put("startSymbol", this.startSymbol.getName());
        }

        if (this.rulesByNonTerminal != null) { // EqualsVerifier: Non-nullity
            final Map<String, Object> object2 = new HashMap<String, Object>();

            for (final NonTerminal nonTerminal : this.rulesByNonTerminal.keySet()) {
                final List<Object> object3 = new ArrayList<Object>();

                for (final Rule rule : this.rulesByNonTerminal.get(nonTerminal)) {
                    object3.add(rule.toObject());
                }

                object2.put(nonTerminal.getName(), object3);
            }

            object1.put("rules", object2);
        }

        return object1;
    }

    /**
     * This method returns the JSON string representation of this context-free
     * grammar.
     *
     * @return The JSON string representation of this context-free grammar.
     */
    @Override
    public String toString() {
        try {
            return Grammar.OBJECT_MAPPER.writeValueAsString(this.toObject());
        } catch (final JsonProcessingException exception) {
            return "";
        }
    }
}
