/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.HashMap;
import java.util.Map;

import net.jcip.annotations.Immutable;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;

/**
 * This abstract class represents a symbol used in a context-free grammar.
 *
 * <p>
 * A symbol can be either a terminal or a non-terminal symbol. All symbols have
 * a string representation that is stored in the name field.
 *
 * <p>
 * <i>For more information see the Wikipedia article: <a
 * href="http://en.wikipedia.org/wiki/Nonterminal">Terminal and nonterminal
 * symbols</a></i>
 *
 * @author Clemens Akens
 */
@Immutable
public abstract class Symbol {

    /**
     * This static field stores the object mapper to generate the JSON string
     * representation from the JSON object representation of a symbol.
     */
    private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

    /** This field stores the string representation of this symbol. */
    private final String name;

    /**
     * This constructor instantiates a new symbol.
     *
     * @param name
     *            The non-empty string representation of the new symbol.
     */
    Symbol(final String name) {
        Preconditions.checkArgument(!name.isEmpty(), "name is empty");

        this.name = name;
    }

    /**
     * This method indicates whether some other object is "equal to" this one.
     *
     * @param object
     *            The other object.
     * @return True, if this object is equal to the other object.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public final boolean equals(final Object object) {
        if (object == this) {
            return true;
        }

        if (object == null) {
            return false;
        }

        if (object.getClass() != this.getClass()) {
            return false;
        }

        final Symbol other = (Symbol) object;

        return Objects.equal(other.name, this.name);
    }

    /**
     * This method returns the string representation of this symbol.
     *
     * @return The string representation of this symbol.
     */
    public final String getName() {
        return this.name;
    }

    /**
     * This method returns the hash code value for this object.
     *
     * @return The hash code value for this object.
     * @see java.lang.Object#hashCode()
     */
    @Override
    public final int hashCode() {
        return Objects.hashCode(this.name);
    }

    /**
     * This method indicates whether this symbol is a non-terminal symbol.
     *
     * @return True, if this symbol is a non-terminal symbol.
     */
    public final boolean isNonTerminal() {
        return this instanceof NonTerminal;
    }

    /**
     * This method indicates whether this symbol is a terminal symbol.
     *
     * @return True, if this symbol is a terminal symbol.
     */
    public final boolean isTerminal() {
        return this instanceof Terminal;
    }

    /**
     * This method returns the JSON object representation of this symbol.
     *
     * @return The JSON object representation of this symbol.
     */
    public final Object toObject() {
        final Map<String, String> object = new HashMap<String, String>();

        if (this.isNonTerminal()) {
            object.put("N", this.name);
        } else {
            object.put("T", this.name);
        }

        return object;
    }

    /**
     * This method returns the JSON string representation of this symbol.
     *
     * @return The JSON string representation of this symbol.
     */
    @Override
    public final String toString() {
        try {
            return Symbol.OBJECT_MAPPER.writeValueAsString(this.toObject());
        } catch (final JsonProcessingException exception) {
            return "";
        }
    }
}
