/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import net.jcip.annotations.Immutable;

/**
 * This class represents a terminal symbol in a context-free grammar.
 *
 * <p>
 * Terminal symbols, or just terminals, are literal characters that can appear
 * in the inputs to or outputs from the production rules of a context-free
 * grammar and that cannot be broken down into "smaller" units. To be precise,
 * terminal symbols cannot be changed using the rules of the grammar. Thus, they
 * are the elementary symbols of the language defined by a context-free grammar.
 *
 * <p>
 * <i>For more information see the Wikipedia article: <a
 * href="http://en.wikipedia.org/wiki/Nonterminal">Terminal and nonterminal
 * symbols</a></i>
 *
 * @author Clemens Akens
 */
@Immutable
public final class Terminal extends Symbol {

    /**
     * This static method creates a new terminal symbol.
     *
     * @param name
     *            The non-empty string representation of the new terminal
     *            symbol.
     * @return The new terminal symbol.
     */
    public static Terminal create(final String name) {
        return new Terminal(name);
    }

    /**
     * This constructor instantiates a new terminal symbol.
     *
     * @param name
     *            The non-empty string representation of the new terminal
     *            symbol.
     */
    private Terminal(final String name) {
        super(name);
    }
}
