/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import net.jcip.annotations.Immutable;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Multimap;

/**
 * This class provides static utility methods for a context-free grammar.
 *
 * @author Clemens Akens
 */
@Immutable
public final class Utils {

    /**
     * This static method returns the direct-left-corner relation of the
     * specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol B is a direct left corner of a non-terminal symbol
     * A, if there is an A-production rule with B as the left-most symbol on the
     * right-hand side.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The direct-left-corner relation of the specified context-free
     *         grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeDirectLeftCornerRelation(final Grammar grammar) {
        final Multimap<NonTerminal, NonTerminal> directLeftCornerRelation = HashMultimap.create();

        for (final Rule rule : grammar.getRules()) {
            if (!rule.getRightHandSide().isEmpty()) {
                final Symbol leftMostSymbol = rule.getRightHandSide().get(0);

                if (leftMostSymbol.isNonTerminal()) {
                    directLeftCornerRelation.put(rule.getLeftHandSide(), (NonTerminal) leftMostSymbol);
                }
            }
        }

        return ImmutableSetMultimap.copyOf(directLeftCornerRelation);
    }

    /**
     * This static method returns the set of all directly left recursive
     * non-terminal symbols of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is directly left recursive if it is a direct left
     * corner of itself.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The set of all directly left recursive non-terminal symbols of
     *         the specified context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeDirectlyLeftRecursiveNonTerminals(final Grammar grammar) {
        return Utils.computeDirectlyLeftRecursiveNonTerminals(grammar, Utils.computeDirectLeftCornerRelation(grammar));
    }

    /**
     * This static method returns the set of all directly left recursive
     * non-terminal symbols of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is directly left recursive if it is a direct left
     * corner of itself.
     *
     * @param grammar
     *            The context-free grammar.
     * @param directLeftCornerRelation
     *            The direct-left-corner relation of the specified context-free
     *            grammar.
     * @return The set of all directly left-recursive non-terminal symbols of
     *         the specified context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeDirectlyLeftRecursiveNonTerminals(final Grammar grammar, final ImmutableSetMultimap<NonTerminal, NonTerminal> directLeftCornerRelation) {
        final Set<NonTerminal> directlyLeftRecursiveNonTerminals = new HashSet<NonTerminal>();

        for (final NonTerminal nonTerminal : grammar.getNonTerminals()) {
            if (directLeftCornerRelation.containsEntry(nonTerminal, nonTerminal)) {
                directlyLeftRecursiveNonTerminals.add(nonTerminal);
            }
        }

        return ImmutableSet.copyOf(directlyLeftRecursiveNonTerminals);
    }

    /**
     * This static method returns the direct-reachability relation of the
     * specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol B is direct reachable from a non-terminal symbol A,
     * if there is an A-production rule with B on the right-hand side.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The direct-reachability relation of the specified context-free
     *         grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeDirectReachabilityRelation(final Grammar grammar) {
        final Multimap<NonTerminal, NonTerminal> directReachabilityRelation = HashMultimap.create();

        for (final Rule rule : grammar.getRules()) {
            for (final Symbol symbol : rule.getRightHandSide()) {
                if (symbol.isNonTerminal()) {
                    directReachabilityRelation.put(rule.getLeftHandSide(), (NonTerminal) symbol);
                }
            }
        }

        return ImmutableSetMultimap.copyOf(directReachabilityRelation);
    }

    /**
     * This static method returns the left-corner relation of the specified
     * context-free grammar.
     *
     * <p>
     * The left-corner relation is defined to be the transitive closure of the
     * direct-left-corner relation.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The left-corner relation of the specified context-free grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeLeftCornerRelation(final Grammar grammar) {
        return Utils.computeLeftCornerRelation(grammar, Utils.computeDirectLeftCornerRelation(grammar));
    }

    /**
     * This static method returns the left-corner relation of the specified
     * context-free grammar.
     *
     * <p>
     * The left-corner relation is defined to be the transitive closure of the
     * direct-left-corner relation.
     *
     * @param grammar
     *            The context-free grammar.
     * @param directLeftCornerRelation
     *            The direct-left-corner relation of the specified context-free
     *            grammar.
     * @return The left-corner relation of the specified context-free grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeLeftCornerRelation(final Grammar grammar, final ImmutableSetMultimap<NonTerminal, NonTerminal> directLeftCornerRelation) {
        final Multimap<NonTerminal, NonTerminal> leftCornerRelation = HashMultimap.create(directLeftCornerRelation);
        final ImmutableSet<NonTerminal> nonTerminals = grammar.getNonTerminals();

        for (final NonTerminal kNonTerminal : nonTerminals) {
            for (final NonTerminal iNonTerminal : nonTerminals) {
                for (final NonTerminal jNonTerminal : nonTerminals) {
                    if (leftCornerRelation.containsEntry(iNonTerminal, kNonTerminal) && leftCornerRelation.containsEntry(kNonTerminal, jNonTerminal)) {
                        leftCornerRelation.put(iNonTerminal, jNonTerminal);
                    }
                }
            }
        }

        return ImmutableSetMultimap.copyOf(leftCornerRelation);
    }

    /**
     * This static method returns the set of all left-recursive non-terminal
     * symbols of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is left recursive if it is a left corner of itself.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The set of all left-recursive non-terminal symbols of the
     *         specified context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeLeftRecursiveNonTerminals(final Grammar grammar) {
        return Utils.computeLeftRecursiveNonTerminals(grammar, Utils.computeLeftCornerRelation(grammar));
    }

    /**
     * This static method returns the set of all left-recursive non-terminal
     * symbols of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is left recursive if it is a left corner of itself.
     *
     * @param grammar
     *            The context-free grammar.
     * @param leftCornerRelation
     *            The left-corner relation of the specified context-free
     *            grammar.
     * @return The set of all left-recursive non-terminal symbols of the
     *         specified context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeLeftRecursiveNonTerminals(final Grammar grammar, final ImmutableSetMultimap<NonTerminal, NonTerminal> leftCornerRelation) {
        final Set<NonTerminal> leftRecursiveNonTerminals = new HashSet<NonTerminal>();

        for (final NonTerminal nonTerminal : grammar.getNonTerminals()) {
            if (leftCornerRelation.containsEntry(nonTerminal, nonTerminal)) {
                leftRecursiveNonTerminals.add(nonTerminal);
            }
        }

        return ImmutableSet.copyOf(leftRecursiveNonTerminals);
    }

    /**
     * This static method returns the set of all nullable non-terminal symbols
     * of the specified context-free grammar.
     *
     * <p>
     * A nullable non-terminal symbol is one that can produce empty by the use
     * of one or more production rules.<br>
     * A production rule is empty if its right-hand side is empty or consists
     * only of nullable non-terminal symbols.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The set of all nullable non-terminal symbols of the specified
     *         context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeNullableNonTerminals(final Grammar grammar) {
        final Set<NonTerminal> nonTerminals = new HashSet<NonTerminal>(grammar.getNonTerminals());
        final Set<NonTerminal> nullableNonTerminals = new HashSet<NonTerminal>();

        boolean changed;

        do {
            changed = false;

            for (final Iterator<NonTerminal> iterator = nonTerminals.iterator(); iterator.hasNext();) {
                final NonTerminal nonTerminal = iterator.next();

                for (final Rule rule : grammar.getRulesByNonTerminal().get(nonTerminal)) {
                    boolean nullable = rule.getRightHandSide().isEmpty();

                    for (final Symbol symbol : rule.getRightHandSide()) {
                        nullable = symbol.isNonTerminal() && nullableNonTerminals.contains(symbol);

                        if (!nullable) {
                            break;
                        }
                    }

                    if (nullable) {
                        nullableNonTerminals.add(nonTerminal);
                        iterator.remove();

                        changed = true;

                        break;
                    }
                }
            }
        } while (changed);

        return ImmutableSet.copyOf(nullableNonTerminals);
    }

    /**
     * This static method returns the sets of all productive production rules of
     * the specified context-free grammar mapped by their respective left-hand
     * sides.
     *
     * <p>
     * A production rule is productive if its right-hand side consists of
     * symbols all of which are productive.<br>
     * Terminal symbols are productive since they produce terminals and empty is
     * productive since it produces the empty string.<br>
     * A non-terminal symbol is productive if there is a productive production
     * rule for it.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The sets of all productive production rules of the specified
     *         context-free grammar mapped by their respective left-hand sides.
     */
    public static ImmutableSetMultimap<NonTerminal, Rule> computeProductiveRulesByNonTerminal(final Grammar grammar) {
        final Multimap<NonTerminal, Rule> productiveRulesByNonTerminal = HashMultimap.create();
        final Set<Rule> rules = new HashSet<Rule>(grammar.getRules());

        boolean changed;

        do {
            changed = false;

            for (final Iterator<Rule> iterator = rules.iterator(); iterator.hasNext();) {
                final Rule rule = iterator.next();
                boolean productive = rule.getRightHandSide().isEmpty();

                for (final Symbol symbol : rule.getRightHandSide()) {
                    productive = symbol.isTerminal() || productiveRulesByNonTerminal.containsKey(symbol);

                    if (!productive) {
                        break;
                    }
                }

                if (productive) {
                    productiveRulesByNonTerminal.put(rule.getLeftHandSide(), rule);
                    iterator.remove();

                    changed = true;
                }
            }
        } while (changed);

        return ImmutableSetMultimap.copyOf(productiveRulesByNonTerminal);
    }

    /**
     * This static method returns the reachability relation of the specified
     * context-free grammar.
     *
     * <p>
     * The reachability relation is defined to be the transitive closure of the
     * direct-reachability relation.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The reachability relation of the specified context-free grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeReachabilityRelation(final Grammar grammar) {
        return Utils.computeReachabilityRelation(grammar, Utils.computeDirectReachabilityRelation(grammar));
    }

    /**
     * This static method returns the reachability relation of the specified
     * context-free grammar.
     *
     * <p>
     * The reachability relation is defined to be the transitive closure of the
     * direct-reachability relation.
     *
     * @param grammar
     *            The context-free grammar.
     * @param directReachabilityRelation
     *            The direct-reachability relation of the specified context-free
     *            grammar.
     * @return The reachability relation of the specified context-free grammar.
     */
    public static ImmutableSetMultimap<NonTerminal, NonTerminal> computeReachabilityRelation(final Grammar grammar, final ImmutableSetMultimap<NonTerminal, NonTerminal> directReachabilityRelation) {
        final ImmutableSet<NonTerminal> nonTerminals = grammar.getNonTerminals();
        final Multimap<NonTerminal, NonTerminal> reachabilityRelation = HashMultimap.create(directReachabilityRelation);

        for (final NonTerminal kNonTerminal : nonTerminals) {
            for (final NonTerminal iNonTerminal : nonTerminals) {
                for (final NonTerminal jNonTerminal : nonTerminals) {
                    if (reachabilityRelation.containsEntry(iNonTerminal, kNonTerminal) && reachabilityRelation.containsEntry(kNonTerminal, jNonTerminal)) {
                        reachabilityRelation.put(iNonTerminal, jNonTerminal);
                    }
                }
            }
        }

        return ImmutableSetMultimap.copyOf(reachabilityRelation);
    }

    /**
     * This static method returns the set of all reachable non-terminal symbols
     * of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is called reachable if there exists at least one
     * sentential form, derivable from the start symbol, in which it occurs.
     *
     * @param grammar
     *            The context-free grammar.
     * @return The set of all reachable non-terminal symbols of the specified
     *         context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeReachableNonTerminals(final Grammar grammar) {
        return Utils.computeReachableNonTerminals(grammar, Utils.computeReachabilityRelation(grammar));
    }

    /**
     * This static method returns the set of all reachable non-terminal symbols
     * of the specified context-free grammar.
     *
     * <p>
     * A non-terminal symbol is called reachable if there exists at least one
     * sentential form, derivable from the start symbol, in which it occurs.
     *
     * @param grammar
     *            The context-free grammar.
     * @param reachabilityRelation
     *            The reachability relation of the specified context-free
     *            grammar.
     * @return The set of all reachable non-terminal symbols of the specified
     *         context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeReachableNonTerminals(final Grammar grammar, final ImmutableSetMultimap<NonTerminal, NonTerminal> reachabilityRelation) {
        final Set<NonTerminal> reachableNonTerminals = new HashSet<NonTerminal>();

        reachableNonTerminals.add(grammar.getStartSymbol());
        reachableNonTerminals.addAll(reachabilityRelation.get(grammar.getStartSymbol()));

        return ImmutableSet.copyOf(reachableNonTerminals);
    }

    /**
     * This static method returns the set of all retained non-terminal symbols
     * of the specified context-free grammar.
     *
     * <p>
     * A retained non-terminal symbol is defined to be one of the following:
     *
     * <ul>
     * <li>The start symbol.
     * <li>A non-terminal symbol that occur on the right-hand side of an
     * production rule in other than the left-most position.
     * <li>A non-terminal symbol that occur on the right-hand side of an
     * production rule whose left-hand side is a non-left-recursive nonterminal
     * symbol.
     * </ul>
     *
     * @param grammar
     *            The context-free grammar.
     * @return The set of all retained non-terminal symbols of the specified
     *         context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeRetainedNonTerminals(final Grammar grammar) {
        return Utils.computeRetainedNonTerminals(grammar, Utils.computeLeftRecursiveNonTerminals(grammar));
    }

    /**
     * This static method returns the set of all retained non-terminal symbols
     * of the specified context-free grammar.
     *
     * <p>
     * A retained non-terminal symbol is defined to be one of the following:
     *
     * <ul>
     * <li>The start symbol.
     * <li>A non-terminal symbol that occur on the right-hand side of an
     * production rule in other than the left-most position.
     * <li>A non-terminal symbol that occur on the right-hand side of an
     * production rule whose left-hand side is a non-left-recursive nonterminal
     * symbol.
     * </ul>
     *
     * @param grammar
     *            The context-free grammar.
     * @param leftRecursiveNonTerminals
     *            The left-recursive non-terminal symbols of the specified
     *            context-free grammar.
     * @return The set of all retained non-terminal symbols of the specified
     *         context-free grammar.
     */
    public static ImmutableSet<NonTerminal> computeRetainedNonTerminals(final Grammar grammar, final ImmutableSet<NonTerminal> leftRecursiveNonTerminals) {
        final Set<NonTerminal> retainedNonTerminals = new HashSet<NonTerminal>();

        retainedNonTerminals.add(grammar.getStartSymbol());

        for (final Rule rule : grammar.getRules()) {
            final ImmutableList<? extends Symbol> symbols = rule.getRightHandSide();

            if (leftRecursiveNonTerminals.contains(rule.getLeftHandSide())) {
                if (symbols.size() > 1) {
                    for (final Symbol symbol : symbols.subList(1, symbols.size())) {
                        if (symbol.isNonTerminal()) {
                            retainedNonTerminals.add((NonTerminal) symbol);
                        }
                    }
                }
            } else {
                for (final Symbol symbol : symbols) {
                    if (symbol.isNonTerminal()) {
                        retainedNonTerminals.add((NonTerminal) symbol);
                    }
                }
            }
        }

        return ImmutableSet.copyOf(retainedNonTerminals);
    }

    /**
     * This constructor instantiates a new utility class.
     */
    private Utils() {
    }
}
